from typing import Any

from django.db import models
from edc_visit_tracking.reason_updater import SubjectVisitReasonUpdater

from ..constants import IN_PROGRESS_APPT, MISSED_APPT


class MissedAppointmentModelMixin(models.Model):
    def create_missed_visit_from_appointment(self: Any):
        if self.appt_timing == MISSED_APPT:
            self.visit_model_cls().objects.create_missed_from_appointment(
                appointment=self,
            )

    def update_subject_visit_reason_or_raise(self: Any):
        """Trys to update the subject_visit.reason field, if it
        exists, when appt_timing changes, or raises.
        """
        if (
            self.id
            and self.appt_status == IN_PROGRESS_APPT
            and self.appt_timing
            and self.appt_reason
        ):
            reason_updater = SubjectVisitReasonUpdater(
                appointment=self,
                appt_timing=self.appt_timing,
                appt_reason=self.appt_reason,
            )
            reason_updater.update_or_raise()

    class Meta:
        abstract = True



from datetime import datetime
from typing import Any

from django.conf import settings
from edc_utils import convert_php_dateformat
from edc_visit_schedule.schedule.window import (
    ScheduledVisitWindowError,
    UnScheduledVisitWindowError,
)

UNSCHEDULED_WINDOW_ERROR = "unscheduled_window_error"
SCHEDULED_WINDOW_ERROR = "scheduled_window_error"


class WindowPeriodFormValidatorMixin:
    def validate_appt_datetime_in_window_period(self: Any, appointment, *args) -> None:
        # if not appointment.is_baseline_appt:
        self.datetime_in_window_or_raise(appointment, *args)

    def validate_visit_datetime_in_window_period(self: Any, *args) -> None:
        self.datetime_in_window_or_raise(*args)

    def validate_crf_datetime_in_window_period(self: Any, *args) -> None:
        self.datetime_in_window_or_raise(*args)

    def datetime_in_window_or_raise(
        self,
        appointment: Any,
        proposed_datetime: datetime,
        form_field: str,
    ):
        datestring = convert_php_dateformat(settings.SHORT_DATE_FORMAT)
        appointment.visit_from_schedule.timepoint_datetime = appointment.timepoint_datetime
        lower = appointment.visit_from_schedule.dates.lower.strftime(datestring)
        try:
            appointment.schedule.datetime_in_window(
                timepoint_datetime=appointment.timepoint_datetime,
                dt=proposed_datetime,
                visit_code=appointment.visit_code,
                visit_code_sequence=appointment.visit_code_sequence,
                baseline_timepoint_datetime=self.baseline_timepoint_datetime(appointment),
            )
        except UnScheduledVisitWindowError:
            upper = appointment.schedule.visits.next(
                appointment.visit_code
            ).dates.lower.strftime(datestring)
            self.raise_validation_error(
                {form_field: f"Invalid. Expected a date between {lower} and {upper} (U)."},
                UNSCHEDULED_WINDOW_ERROR,
            )
        except ScheduledVisitWindowError:
            upper = appointment.visit_from_schedule.dates.upper.strftime(datestring)
            self.raise_validation_error(
                {form_field: f"Invalid. Expected a date between {lower} and {upper} (S)."},
                SCHEDULED_WINDOW_ERROR,
            )

    @staticmethod
    def baseline_timepoint_datetime(appointment) -> datetime:
        return appointment.__class__.objects.first_appointment(
            subject_identifier=appointment.subject_identifier,
            visit_schedule_name=appointment.visit_schedule_name,
            schedule_name=appointment.schedule_name,
        ).timepoint_datetime
